# frozen_string_literal: false
#
#   irb/version.rb - irb version definition file
#   	$Release Version: 0.9.6$
#   	$Revision$
#   	by Keiju ISHITSUKA(keiju@ishitsuka.com)
#
# --
#
#
#

module IRB # :nodoc:
  VERSION = "1.4.1"
  @RELEASE_VERSION = VERSION
  @LAST_UPDATE_DATE = "2021-12-25"
end
